/*
 * B-H_DiceTower_v2.c
 *
 * Created: 1/19/2020 2:03:41 PM
 * Author : johna
 *
 * Copyright (C) 2020 by John Anderson <racerxr650r@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any 
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES 
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN 
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */ 

#include <avr/io.h>
#include <stdlib.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay_basic.h>

// Program constants **********************************************************
#define BLANK_OUTPUT				00
#define DISPLAY_BLANK_TIME			100
#define DISPLAY_CONFIG_TIME			500

// Busy Delay function ********************************************************
#define F_CPU_KHZ 8000  // CPU clock speed in kHz
#define LOOPS_PER_MS (uint16_t)(F_CPU_KHZ/4)

// Program data types *********************************************************
typedef enum {State_Init, State_Idle, State_Update_Count, State_Display_Count, State_Update_Dice, State_Display_Dice, State_Roll, State_Roll_Result}State;

// Busy wait delay function ***************************************************
void delay_ms(uint16_t ms)
{
	while(ms--)
	_delay_loop_2(LOOPS_PER_MS);
}

// I/O Constants **************************************************************
/* Port B Usage ******************************
Port	Config		Device
PB0		DI			Thumbwheel Digit 1 Bit 3
PB1		DO			LED
PB2		DI			Button
PB3		OC2A		Speaker
PB4		MISO		ISP-MISO
PB5		SCK			ISP-SCK
PB6		DI			Thumbwheel Digit 2 Bit 1
PB7		DI			Thumbwheel Digit 2 Bit 0
*********************************************/
#define TW_D1_B3	0b00000001
#define LED			0b00000010
#define BUTTON		0b00000100
#define SPEAKER		0b00001000
#define ISP_MISO	0b00010000
#define ISP_SCK		0b00100000
#define TW_D2_B1	0b01000000
#define TW_D2_B0	0b10000000
/* Port C Usage ******************************
Port	Config		Device
PC0		DI			Thumbwheel Digit 1 Bit 0
PC1		DI			Thumbwheel Digit 1 Bit 1
PC2		DI			Thumbwheel Digit 1 Bit 2
PC3		DI			Rotary Bit 0
PC4		DI			Rotary Bit 1
PC5		DI			Rotary Bit 2
PC6		RESET		ISP-RESET
*********************************************/
#define TW_D1_B0	0b00000001
#define TW_D1_B1	0b00000010
#define TW_D1_B2	0b00000100
#define ROT_B0		0b00001000
#define ROT_B1		0b00010000
#define ROT_B2		0b00100000
#define ISP_RESET	0b01000000
/* Port C Usage ******************************
Port	Config		Device
PD0		D0			Nixie Digit 1 Bit 0
PD1		D0			Nixie Digit 1 Bit 1
PD2		D0			Nixie Digit 1 Bit 2
PD3		D0			Nixie Digit 1 Bit 3
PD4		D0			Nixie Digit 2 Bit 0
PD5		D0			Nixie Digit 2 Bit 1
PD6		D0			Nixie Digit 2 Bit 2
PD7		D0			Nixie Digit 2 Bit 3
*********************************************/
#define NIX_D1_B0	0b00000001
#define NIX_D1_B1	0b00000010
#define NIX_D1_B2	0b00000100
#define NIX_D1_B3	0b00001000
#define NIX_D2_B0	0b00010000
#define NIX_D2_B1	0b00100000
#define NIX_D2_B2	0b01000000
#define NIX_D2_B3	0b10000000

// Setup the IO pins **********************************************************
void initIO()
{
	// Set appropriate Port B pins output (the rest are input or alt function)
	DDRB = LED;
	// Enable pull ups on port B
	PORTB = BUTTON | TW_D1_B3 | TW_D2_B0 | TW_D2_B1;

	// Set Port C pins to input (no outputs)
	DDRC = ~(TW_D1_B0 | TW_D1_B1 | TW_D1_B2 | ROT_B0 | ROT_B1 | ROT_B2);
	// Enable pull ups on Port C
	PORTC = TW_D1_B0 | TW_D1_B1 | TW_D1_B2 | ROT_B0 | ROT_B1 | ROT_B2;

	// Set Port D pins to output
	DDRD = NIX_D1_B0 | NIX_D1_B1 | NIX_D1_B2 | NIX_D1_B3 | NIX_D2_B0 | NIX_D2_B1 | NIX_D2_B2 | NIX_D2_B3;
}

// I/O interface functions ****************************************************
// Set the status LED *********************************************************
void led(uint8_t state)
{
	if(state)
		PORTB |= LED;
	else
		PORTB &= ~LED;
}

// Output a two digit decimal number on the Nixie Tube display ****************
void output(int out)
{
	// Calculate the ones digit
	int outputByte = out%10;
	// Calculate the tens digit
	out /= 10;
	outputByte |= (out%10)<<4;
	// Write to the output port
	PORTD = outputByte;
}

// Get the dice count from the thumbwheel input *******************************
int getCount()
{
	// Arrange the bits from port B & C for the ones digit
	int count = (~PINC & (TW_D1_B0 | TW_D1_B1 | TW_D1_B2)) | ((~PINB & TW_D1_B3) << 3);
	// Arrange the bits from port B for the tens digit, multiply it times 10, and
	// add it to the count, NOTE: Bit 0 and 1 are flipped on the port
	count += ((((~PINB & TW_D2_B1)>>5)|((~PINB & TW_D2_B0)>>7))*10);
	
	return(count);
}

// Get the dice type **********************************************************
const int DieSides[] = {2,4,6,8,10,12,20,100};
int getDice()
{
	return(DieSides[(~PINC & (ROT_B0 | ROT_B1 |ROT_B2)) >> 3]);
}

// Get the status of the roll button ******************************************
uint8_t getButton()
{
	uint8_t ret = ((~PINB)&BUTTON)>>2;
	return(ret);
}

// Sound output functions *****************************************************
// Include PCM sound images ***************************************************
#include "dt_numbers_7.h"
// Sound globals, handlers, and functions *************************************
#define IS_SOUND_PLAYING gTmrCntr

#define SND_HIGH	0x03
#define SND_MEDIUM	0x07
#define SND_LOW		0x0f

#define TONE_C8		4186
#define TONE_B7		3951
#define TONE_As7	3729
#define TONE_A7		3520
#define TONE_Gs7	3322
#define TONE_G7		3136
#define TONE_Fs7	2960
#define TONE_F7		2794
#define TONE_E7		2637
#define TONE_Ds7	2489
#define TONE_D7		2349
#define TONE_Cs7	2217
#define TONE_C7		2093
#define TONE_B6		1976
#define TONE_As6	1865
#define TONE_A6		1760
#define TONE_Gs6	1661
#define TONE_G6		1568
#define TONE_F6		1397
#define TONE_E6		1319
#define TONE_Ds6	1245
#define TONE_D6		1175
#define TONE_Cs6	1109
#define TONE_C6		1047
#define TONE_B5		988
#define TONE_As5	932
#define TONE_A5		880
#define TONE_Gs5	831
#define TONE_G5		784
#define TONE_Fs5	740
#define TONE_F5		698
#define TONE_E5		659
#define TONE_Ds5	622
#define TONE_D5		587
#define TONE_Cs5	554
#define TONE_C5		523
#define TONE_B4		494
#define TONE_As4	466
#define TONE_A4		440
#define TONE_Gs4	415
#define TONE_G4		392
#define TONE_Fs4	370
#define TONE_F4		349
#define TONE_E4		330
#define TONE_Ds4	311
#define TONE_D4		294
#define TONE_Cs4	277
#define TONE_C4		262
#define TONE_B3		247
#define TONE_As3	233
#define TONE_A3		220
#define TONE_Gs3	208
#define TONE_G3		196
#define TONE_Fs3	185
#define TONE_F3		175
#define TONE_E3		165
#define TONE_Ds3	156
#define TONE_D3		147
#define TONE_Cs3	139
#define TONE_C3		131
#define TONE_B2		123
#define TONE_As2	117
#define TONE_A2		110
#define TONE_Gs2	104
#define TONE_G2		98
#define TONE_Fs2	92
#define TONE_F2		87
#define TONE_E2		82
#define TONE_Ds2	78
#define TONE_D2		74
#define TONE_Cs2	69
#define TONE_C2		65
#define TONE_B1		62
#define TONE_As1	58
#define TONE_A1		55
#define TONE_Gs1	52
#define TONE_G1		49
#define TONE_Fs1	46
#define TONE_F1		44
#define TONE_E1		41
#define TONE_Ds1	39
#define TONE_D1		37
#define TONE_Cs1	35
#define TONE_C1		33

volatile uint32_t gTmrCntr, gTmrEnd;
volatile uint16_t gSndCnt, gSndEnd, gRleCnt;
volatile uint8_t  gRate;
volatile const uint8_t  *gSound;

// Load the PWM output value
void inline loadPWM(void)
{
	uint8_t byte;
	
	if(gRleCnt)
	{
		--gRleCnt;
		OCR2A = 0x80;
	}
	else
	{
		byte = pgm_read_byte_near(&gSound[gSndCnt++]);
		if(byte)
			OCR2A = byte;
		else
		{
			gRleCnt = pgm_read_byte_near(&gSound[gSndCnt++])-1;
			OCR2A = 0x80;
		}
	}
}

// Enable the PWM/Timer and overflow interrupt
void startPWM(uint16_t freq)
{
	// Zero the Hardware Timer Counter
	TCNT2 = 0x00;
	// Zero the Software Timer Counter
	gTmrCntr = 1;
	// Set the DDR for the output pin
	DDRB |= SPEAKER;
	// If no frequency provided, play a PCM'd sound from flash
	if(!freq)
	{
		// Set the run length encoded counter to zero
		gRleCnt = 0;
		// Set up the Compare Output and Wave Generation modes
		// non-inverting, PWM fast mode
		TCCR2A = 0x83;
		// Load the first PCM value into the PWM output comparator
		loadPWM();
		// Enable the overflow interrupt on timer 0
		TIMSK2 = (1<<TOIE2);
		// Start the timer
		// No pre-scaler
		// Timer frequency = 8 mHz system clock
		TCCR2B = 0x01;
	}
	else
	{
		
		// Set up the Compare Output and Wave Generation modes
		// Toggle OC0A, CTC mode
		TCCR2A = 0x42;
		// Calculate the output compare
		// Divide the timer frequency by 2 because the full period requires to compare/overflows
		OCR2A = (uint8_t)((uint16_t)31250/(uint16_t)freq);
		// Enable the timer 0 output compare A interrupt
		TIMSK2 = 0x02;
		// Start the timer
		// 256 pre-scaler
		// Timer frequency = 8 mHz system clock / 128 prescaler = 62500 Hz
		TCCR2B = 0x05;
	}
}

// Disable the PWM/Timer and overflow interrupt
void stopPWM(void)
{
	// Stop the timer
	TCCR2B = 0;
	// Clear the Compare Output and Wave Generation modes
	TCCR2A = 0;
	// Clear the DDR for the output pin
	DDRB ^= SPEAKER;
	// Disable the overflow interrupt on timer 0
	TIMSK2 = 0;
	// Zero the Sound Counter so a new sound can be played
	gSndCnt = 0;
	// Zero the Software Timer Counter
	gTmrCntr = 0;
}

// Interrupt handler for PCM'd sound output (Timer 2 overflow)
ISR(TIMER2_OVF_vect)
{
	++gTmrCntr;
	
	// Update according to sample rate (High - 8K, Medium - 4K, Low - 2K)
	if(!(gTmrCntr & gRate))
	{
		// Load the PWM compare register with the next PCM value
		loadPWM();
		
		// If this is the end of the sound data, stop the PWM/Timer
		if((gSndCnt > gSndEnd) && (!gRleCnt))
			stopPWM();
	}
}

// Interrupt handler for tone output (Timer 0 Output Compare A)
ISR(TIMER2_COMPA_vect)
{
	++gTmrCntr;
	
	// If this is the end of the tone, stop the PWM/Timer
	if(gTmrCntr > gTmrEnd)
		stopPWM();
}

static int8_t playSound(const uint8_t *sound, uint8_t rate, uint16_t end)
{
	int8_t ret = 0;
	
	// If no sound is currently playing
	if(!IS_SOUND_PLAYING)
	{
		gRate = rate;
		gSound = sound;
		gSndEnd = end;
		
		// Start the PWM/Timer and interrupts
		startPWM(0);
	}
	// There is a sound playing so return ERROR
	else
	ret = -1;
	
	return(ret);
}

static int8_t playSoundWait(const uint8_t *sound, uint8_t rate, uint16_t end)
{
	int8_t ret;
	
	ret = playSound(sound,rate,end);
	while(IS_SOUND_PLAYING);
	
	return(ret);
}

static int8_t playTone(uint16_t freq, uint16_t dur)
{
	int8_t ret = 0;
	
	if(!gTmrCntr)
	{
		gTmrEnd = ((uint32_t)freq * (uint32_t)dur / 500);
		startPWM(freq);
	}
	else
	ret = -1;
	
	return(ret);
}

static int8_t playToneWait(uint16_t freq, uint16_t dur,uint8_t delay)
{
	int8_t ret;
	
	// Setup and play the tone
	ret = playTone(freq,dur);
	// busy wait until tone is complete
	while(IS_SOUND_PLAYING);
	
	delay_ms(delay);
	
	return(ret);
}

#define GLISSANDO_STEP 10

static int8_t playToneGlissando(int start, int end, int dur)
{
	int      steps = dur/GLISSANDO_STEP, remainder = dur%GLISSANDO_STEP;
	int      i, delta = (end-start)/steps;
	int8_t	 ret;
	
	for(i=steps;i;--i)
	{
		ret = playToneWait(start,GLISSANDO_STEP,0);
		start = start + delta;
	}
	if(remainder)
	ret = playToneWait(end,remainder,0);
	
	return(ret);
}

const uint8_t * onesPCM[] = {NULL,one,two,three,four,five,six,seven,eight,nine,ten,eleven,twelve,thirteen,fourteen,fifteen,sixteen,seventeen,eighteen,nineteen,twenty};
const unsigned onesSize[] = {0,sizeof(one),sizeof(two),sizeof(three),sizeof(four),sizeof(five),sizeof(six),sizeof(seven),sizeof(eight),sizeof(nine),sizeof(ten),sizeof(eleven),sizeof(twelve),sizeof(thirteen),sizeof(fourteen),sizeof(fifteen),sizeof(sixteen),sizeof(seventeen),sizeof(eighteen),sizeof(nineteen),sizeof(twenty)};
const uint8_t * tensPCM[] = {NULL,NULL,twenty,thirty,forty,fifty,sixty,seventy,eighty,ninety};
const unsigned tensSize[] = {0,0,sizeof(twenty),sizeof(thirty),sizeof(forty),sizeof(fifty),sizeof(sixty),sizeof(seventy),sizeof(eighty),sizeof(ninety)};

void speak(int num)
{
	int ones = num%10, tens = (num/10)%10;
	
	if(num<100 && num>0)
	{
		if(tens>1)
		{
			playSoundWait(tensPCM[tens],SND_HIGH,tensSize[tens]);
			delay_ms(25);
		}
		else if(tens)
			playSoundWait(onesPCM[num],SND_HIGH,onesSize[num]);
		
		if((tens!=1) && ones)
			playSoundWait(onesPCM[ones],SND_HIGH,onesSize[ones]);
	}
	
	delay_ms(25);
}

// Program Entry Point ********************************************************
//*****************************************************************************
int main(void)
{
	unsigned int seed = 0, seeded = 0;
	int          i, count = 0, dice = 0, result = 0, timer = 0;
	State        state = State_Init;
	
	initIO();
	// Enable interrupts ******************************************************
	sei();
	
	// Forever
	while (1)
	{
		// Increment the seed counter
		++seed;
		
		// Implement the states of the application
		switch(state)
		{
			// Initialization
			case State_Init:
				count = getCount();
				dice = getDice();
				result = 0;
				timer = 0;
				state = State_Idle;
				break;
			// Idle
			case State_Idle:
				output(result);
				if(getDice() != dice)
				{
					timer = DISPLAY_BLANK_TIME;
					state = State_Update_Dice;
					led(0);
				}
				else if(getCount() != count)
				{
					timer = DISPLAY_BLANK_TIME;
					state = State_Update_Count;
					led(0);
				}
				else if(getButton())	
				{
					output(0);
					led(0);
					speak(getCount());
					delay_ms(250);
					playSoundWait(D,SND_HIGH,sizeof(D));
					delay_ms(250);
					if(getDice()<100)
						speak(getDice());
					else
						speak(99);
					delay_ms(50);
//					There's not enough room in flash for the rolling sound :-(
//					playSound(rolling,SND_HIGH,sizeof(rolling));
					timer = 10;
					state = State_Roll;
				}
				break;
			// Set the dice type to be rolled
			case State_Update_Dice:
				output(BLANK_OUTPUT);
				if(!--timer)
				{
					dice = getDice();
					timer = DISPLAY_CONFIG_TIME;
					result = 0;
					state = State_Display_Dice;
				}
				break;
			// Momentarily display the dice type
			case State_Display_Dice:
				if(dice == 100)
					output(99);
				else
					output(dice);
				if(getDice() != dice)
				{
					timer = DISPLAY_BLANK_TIME;
					state = State_Update_Dice;
				}
				else if(!--timer)
					state = State_Idle;
				break;
			// Set the number of dice to be rolled
			case State_Update_Count:
				output(BLANK_OUTPUT);
				if(!--timer)
				{
					count = getCount();
					timer = DISPLAY_CONFIG_TIME;
					result = 0;
					state = State_Display_Count;
				}
				break;
			// Momentarily display the dice count
			case State_Display_Count:
				output(count);
				if(getCount() != count)
				{
					timer = DISPLAY_BLANK_TIME;
					state = State_Update_Count;
				}
				else if(!--timer)
					state = State_Idle;
				break;
			// Simulate a roll by flashing random values of the dice type on the display
			case State_Roll:
				output((rand()%dice)+1);
				if(!--timer)
					state = State_Roll_Result;
				break;
			// Calculate and display the roll result
			case State_Roll_Result:
				// If not seeded...
				if(!seeded)
				{
					// Seed the rand() function with the seed counter
					// that has been incrementing since startup
					srand(seed);
					seeded = 1;
				}
				result = 0;
				for(i=0;i<count;++i)
					result+=(rand()%dice)+1;

				output(result);
				led(1);
				speak(result);
				state = State_Idle;
				break;
			
		}
		// Insert delay based on the current state
		// Note: if init or idle state, no delay inserted and
		// The seed counter increments as fast as this loop spins.
		if(state==State_Roll)
			delay_ms(100);
		else if(state>State_Idle)
			delay_ms(1);
	}
}

